from multiprocessing.shared_memory import SharedMemory
import numpy as np
from multiprocessing.managers import SharedMemoryManager
from numpy._typing import DTypeLike, _ShapeLike  # noqa
from typing import NamedTuple, Iterable, Callable, Any


class SharedArray(NamedTuple):
    """
    A pointer to a numpy.ndarray in shared memory
    For details, see https://docs.python.org/3/library/multiprocessing.shared_memory.html

    Example that sums a matrix by summing each row in parallel:
    - create a SharedMemoryManager, otherwise the shared memory would not be freed
    - create a shared copy of the array SharedArray.copy(manager, x)
    - pool workers unpack the shared array with the as_array() method
    - use pool.imap to get a generator for tdqm

    ```
    def sum_row(sai):
      sa, i = sai
      x = sa.as_array()
      return x[i].sum()

    def sum_by_row():
        x = np.ones([1000, 1000])
        with SharedMemoryManager() as manager:
            snp = SharedArray.copy(manager, x)
            with Pool() as pool:
                result_generator = pool.imap(sum_row, ((snp, i) for i in range(1000)))
                result_generator = tqdm(result_generator, total=1000, position=0, leave=True)
                return sum(result_generator)
    ```
    """
    shm: SharedMemory
    shape: _ShapeLike
    dtype: DTypeLike

    def as_array(self) -> np.ndarray:
        return np.ndarray(self.shape, dtype=self.dtype, buffer=self.shm.buf)

    @staticmethod
    def allocate_like(manager: SharedMemoryManager, array: np.ndarray):
        shm = manager.SharedMemory(array.nbytes)
        return SharedArray(shm=shm, shape=array.shape, dtype=array.dtype)

    @staticmethod
    def allocate(manager: SharedMemoryManager, shape: _ShapeLike, dtype: DTypeLike, nbytes=None):
        nbytes = nbytes or np.zeros(1, dtype=dtype).nbytes
        shm = manager.SharedMemory(np.prod(shape) * nbytes) # noqa
        return SharedArray(shm=shm, shape=shape, dtype=dtype)

    @staticmethod
    def copy(manager: SharedMemoryManager, array: np.ndarray):
        sa = SharedArray.allocate_like(manager, array)
        a = sa.as_array()
        a[:] = array[:]
        return sa


class PartialStarApply:
    """
    wraps a function call to a function of np.array f(x1, x2, ..., arg1, arg2, ...)
    where the arrays x1, x2, ... are SharedArray and args are passed as a tuple (arg1, arg2, ...)

    Intended for use with multithreading.Pool.imap
    """

    __slots__ = "func", "args", "__dict__", "__weakref__"
    func: Callable[[Iterable], Any]
    args: Iterable[SharedArray]

    def unpacked_args(self) -> list[SharedArray]:
        return [a.as_array() if isinstance(a, SharedArray) else a for a in self.args]

    def __init__(self, func: Callable[..., Any], *args: SharedArray):
        assert isinstance(func, Callable)
        for a in args:
            assert isinstance(a, SharedArray)
        self.func = func
        self.args = args

    def __call__(self, args: Iterable):
        return self.func(*self.unpacked_args(), *args)


import logging
from typing import Union
from typing import Optional

from scipy.stats import rankdata
import numpy as np
import pandas as pd

from gene_trajectory.diffusion_map import diffusion_map, get_symmetrized_affinity_matrix
from gene_trajectory.util.input_validation import validate_matrix

logger = logging.getLogger()


def get_gene_embedding(
        dist_mat: np.array,
        k: int = 10,
        sigma: Union[float, np.array, list] = None,
        n_ev: int = 30,
        t: int = 1,
) -> tuple[np.array, np.array]:
    """
    Get the diffusion embedding of genes based on the gene-gene Wasserstein distance matrix

    :param dist_mat: dist.mat matrix; Gene-gene Wasserstein distance matrix (symmetric)
    :param k: Adaptive kernel bandwidth for each point set to be the distance to its `K`-th nearest neighbor
    :param sigma: Fixed kernel bandwidth, `sigma` will be ignored if `K` is specified
    :param n_ev: Number of leading eigenvectors to export
    :param t: Number of diffusion times
    :return: the diffusion embedding and the eigenvalues
    """
    validate_matrix(dist_mat, square=True)

    k = min(k, dist_mat.shape[0])
    n_ev = min(n_ev + 1, dist_mat.shape[0])
    diffu_emb, eigen_vals = diffusion_map(dist_mat=dist_mat, k=k, sigma=sigma, n_ev=n_ev, t=t)
    diffu_emb = diffu_emb[:, 1:n_ev + 1]
    eigen_vals = eigen_vals[1:n_ev + 1]
    return diffu_emb, eigen_vals


def get_randow_walk_matrix(
        dist_mat: np.array,
        k: int = 10,
) -> np.array:
    """
    Convert a distance matrix into a random-walk matrix based on adaptive Gaussian kernel

    :param dist_mat: Precomputed distance matrix (symmetric)
    :param k: Adaptive kernel bandwidth for each point set to be the distance to its `K`-th nearest neighbor
    :return: Random-walk matrix
    """
    validate_matrix(dist_mat, square=True)

    affinity_matrix_symm = get_symmetrized_affinity_matrix(dist_mat=dist_mat, k=k)
    normalized_vec = 1 / affinity_matrix_symm.sum(axis=1)
    affinity_matrix_norm = (affinity_matrix_symm * normalized_vec[:, None])

    return affinity_matrix_norm


def get_gene_pseudoorder(
        dist_mat: np.array,
        subset: list[int],
        max_id: Optional[int] = None,
) -> np.array:
    """
    Order genes along a given trajectory

    :param dist_mat: Gene-gene Wasserstein distance matrix (symmetric)
    :param subset: Genes in a given trajectory
    :param max_id: Index of the terminal gene
    :return: The pseudoorder
    """
    validate_matrix(dist_mat, square=True)

    emd = dist_mat[subset][:, subset]
    dm_emb, _ = diffusion_map(emd)
    pseudoorder = rankdata(dm_emb[:, 1])

    if max_id is not None and max_id in subset:
        n = len(subset)
        if 2 * pseudoorder[subset.index(max_id)] < n:
            pseudoorder = n + 1 - pseudoorder

    pseudoorder_all = np.zeros(dist_mat.shape[0])
    pseudoorder_all[subset] = pseudoorder
    return pseudoorder_all


def extract_gene_trajectory(
        gene_embedding: pd.DataFrame,
        dist_mat: np.array,
        gene_names: list[str],
        t_list: Union[float, np.array, list],
        n: Optional[int] = None,
        dims=5,
        k=10,
        quantile=0.02,
        other: str = 'Other',
) -> pd.DataFrame:
    """
    Extract gene trajectories

    :param gene_embedding: Gene embedding
    :param dist_mat: Gene-gene Wasserstein distance matrix (symmetric)
    :param gene_names:
    :param t_list:  Number of diffusion times to retrieve each trajectory
    :param n: Number of gene trajectories to retrieve. Will be set to the length of t_list
    :param dims: Dimensions of gene embedding to use to identify terminal genes (extrema)
    :param k: Adaptive kernel bandwidth for each point set to be the distance to its `K`-th nearest neighbor
    :param quantile: Thresholding parameter to extract genes for each trajectory. Default: 0.02
    :param other: Label for genes not in a trajectory. Default: 'Other'
    :return: A data frame indicating gene trajectories and gene ordering along each trajectory
    """
    validate_matrix(dist_mat, square=True)

    if np.isscalar(t_list):
        if n is None:
            raise ValueError(f'n should be specified if t_list is a number: {t_list}')
        t_list = np.full(n, t_list)
    elif n is None:
        n = len(t_list)
    if n != len(t_list):
        raise ValueError(f't_list ({t_list}) should have the same dimension as n ({n})')

    dist_to_origin = np.sqrt((gene_embedding[:, :dims] ** 2).sum(axis=1))
    df = pd.DataFrame(gene_embedding[:, :dims], columns=[f'DM_{i + 1}' for i in range(dims)],
                      index=gene_names).assign(selected=other)
    n_genes = gene_embedding.shape[0]

    diffusion_mat = get_randow_walk_matrix(dist_mat, k=k)

    for i in range(n):
        if sum(df.selected == other) == 0:
            logger.warning(f"Early stop reached. {i - 1} gene trajectories were retrieved.")
            break

        dist_to_origin[df.selected != other] = -np.infty

        seed_idx = np.argmax(dist_to_origin)
        logger.info(f'Generating trajectory from {gene_names[seed_idx]}')

        seed_diffused = np.zeros(n_genes)
        seed_diffused[seed_idx] = 1

        for _ in range(t_list[i]):
            seed_diffused = diffusion_mat @ seed_diffused

        cutoff = max(seed_diffused) * quantile

        trajectory_label = f'Trajectory-{i + 1}'
        df.loc[(seed_diffused > cutoff) & (df.selected == other), 'selected'] = trajectory_label

        df[f'Pseudoorder-{i + 1}'] = get_gene_pseudoorder(dist_mat=dist_mat,
                                                          subset=list(np.where(df.selected == trajectory_label)[0]),
                                                          max_id=seed_idx)
    return df



import logging
import numpy as np
import scanpy as sc
from sklearn.neighbors import NearestNeighbors
import igraph as ig

logger = logging.getLogger()


def get_graph_distance(
        adata: sc.AnnData,
        reduction: str = 'X_dm',
        k: int = 10,
        dims: int = 5,
) -> np.array:
    """
    Compute the graph distance on a Scanpy object

    :param adata: a scanpy Anndata object
    :param reduction: dimensionality reduction to use, default: 'X_dm'
    :param k: adaptive kernel bandwidth for each point set to be the distance to its k-th nearest neighbor
    :param dims: the dimensions to use as input features for kNN graph construction
    :return: a cell-cell graph distance matrix
    """
    # Implementation note. This could also be done using scipy, although it's 3x slower
    #
    # from scipy.sparse.csgraph import shortest_path
    # ds = scipy.sparse.csgraph.shortest_path(adj, directed=False, unweighted=True)
    #
    cell_embedding = adata.obsm[reduction][:, :dims]
    knn = NearestNeighbors(n_neighbors=k + 1).fit(cell_embedding)
    neighbors = knn.kneighbors(cell_embedding, return_distance=False)
    g = ig.Graph(((nn[0], j) for nn in neighbors for j in nn[1:]), directed=False)
    if not g.is_connected():
        raise RuntimeError('The cell-cell kNN graph has disconnected components. Please increase k.')
    graph_dist_mat = np.array(g.distances())
    logger.info("The largest graph distance is", graph_dist_mat.max())

    return graph_dist_mat



from typing import Optional

import numpy as np
import scanpy as sc
from sklearn.cluster import KMeans

from gene_trajectory.util.input_validation import validate_matrix


def select_top_genes(
        adata: sc.AnnData,
        layer: str = None,
        min_expr_percent: float = 0.01,
        max_expr_percent: float = 0.5,
        n_variable_genes: int = 2000,
) -> np.ndarray:
    """
    Narrow down the gene list for gene-gene distance computation by focusing on the top
    2000 variable genes expressed by 1% - 50% of cells.

    :param adata: a scanpy Anndata object
    :param layer: the layer with count data (e.g. 'counts', which can be created
           as `adata.layers['counts']=adata.raw.X.copy()`)
    :param min_expr_percent: minimum fraction of cells expressing the gene
    :param max_expr_percent: maximum fraction of cells expressing the gene
    :param n_variable_genes: number of variable genes to use (default: 2000)
    :return: the cell-cell graph distance matrix
    """
    if layer not in adata.layers.keys():
        raise ValueError(f'Layer {layer} not found in adata. Available {list(adata.layers.keys())}')
    if adata.n_vars < 2000:
        raise ValueError(f'This method should be run with at least 2000 features (genes). The data has {adata.n_vars}')

    sc.pp.calculate_qc_metrics(adata, layer=layer, inplace=True)
    sc.pp.highly_variable_genes(adata, layer=layer, n_top_genes=n_variable_genes, flavor='seurat_v3', inplace=True)
    expr_percent = adata.var['n_cells_by_counts'] / adata.n_obs
    genes = adata.var_names[adata.var['highly_variable'] &
                            (expr_percent > min_expr_percent) & (expr_percent < max_expr_percent)]
    genes_sorted = adata[:, genes].var.sort_values('variances_norm', ascending=False).index
    return genes_sorted.values

# Implementation notes
# - It would look simpler to do `np.count_nonzero(adata.layer[''], axis=0)` instead of running
#   calculate_qc_metrics and getting `n_cells_by_counts`.
#   Unfortunately one needs different implementations for dense/sparse matrices or to convert using
#   - adata.to_df(layer)
#   - np.asarray(np.sum(XX, axis=0).squeeze())


def coarse_grain(
        cell_embedding: np.ndarray,
        gene_expression: np.ndarray,
        graph_dist: np.ndarray,
        n: int = 1000,
        cluster: Optional[np.array] = None,
        random_seed: int = 1,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Apply coarse-graining to reduce the number of cells

    :param cell_embedding: the cell embedding
    :param gene_expression: the gene expression matrix
    :param graph_dist: the graph distance matrix
    :param n: number of cells to keep
    :param cluster: specify an array to use precomputed clusters. If not specified a KMeans clustering will be performed
    :param random_seed: the random seed
    :return: the updated cell embedding and gene expression matrices
    """
    validate_matrix(gene_expression, obj_name='Gene Expression Matrix', min_value=0)
    ncells, ngenes = gene_expression.shape
    validate_matrix(cell_embedding, obj_name='Cell embedding', nrows=ncells)

    if cluster is None:
        k_means = KMeans(n_clusters=n, random_state=random_seed).fit(cell_embedding)
        cluster = k_means.labels_ # noqa

    knn_membership = np.zeros((n, cell_embedding.shape[0]))
    for i, c in enumerate(cluster):
        knn_membership[c, i] = 1
    gene_expression_updated = knn_membership @ gene_expression

    knn_membership_norm = knn_membership / np.sum(knn_membership, axis=1)[:, None]
    graph_dist_updated = knn_membership_norm @ graph_dist @ knn_membership_norm.T

    return gene_expression_updated, graph_dist_updated


def coarse_grain_adata(
        adata: sc.AnnData,
        graph_dist: np.array,
        features: list[str],
        n: int = 1000,
        reduction: str = "X_dm",
        dims: int = 5,
        random_seed=1,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Apply coarse-graining to reduce the number of cells

    :param adata: a scanpy Anndata object
    :param graph_dist: the graph distance matrix
    :param features: the features (i.e. genes) to keep
    :param n: number of cells to keep (default = 1000)
    :param reduction: the dimensional reduction (in adata.obsm) to keep
    :param dims: the number of dimensions to keep (default = 5)
    :param random_seed: the random seed
    :return: the updated cell embedding and gene expression matrices
    """
    if reduction not in adata.obsm_keys():
        raise ValueError(f'Reduction "{reduction}" is not present. Available: {adata.obsm_keys()}')

    cell_embedding = adata.obsm[reduction][:, :dims]
    gene_expression = adata[:, features].X
    cg = coarse_grain(cell_embedding=cell_embedding,
                      gene_expression=gene_expression,
                      graph_dist=graph_dist,
                      n=n, random_seed=random_seed)
    return cg



from typing import Iterable, Callable
from ipywidgets import VBox, HBox, Widget, IntSlider, interactive_output, Output, Label, Button, FloatSlider, TagsInput, \
    Layout
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt

from gene_trajectory.extract_gene_trajectory import extract_gene_trajectory
from gene_trajectory.plot.gene_trajectory_plots import plot_gene_trajectory_3d


class ExtractGeneTrajectoryWidget(HBox):
    widgets: dict[str, Widget] = None
    t_list_widgets: list[IntSlider] = None
    dims_widget: IntSlider = None
    k_widget: IntSlider = None
    quantile_widget: FloatSlider = None
    label_genes: TagsInput = None

    f: Callable = None
    option_panel: VBox = None
    output: Output = None
    gene_trajectory: pd.DataFrame = None
    max_t: int = 20

    @property
    def dims(self) -> int:
        return self.dims_widget.value  # noqa

    @property
    def k(self) -> int:
        return self.k_widget.value  # noqa

    @property
    def quantile(self) -> float:
        return self.quantile_widget.value  # noqa

    @property
    def tlist(self) -> list[int]:
        return [t.value for t in self.t_list_widgets]  # noqa

    def __init__(
            self,
            gene_embedding: pd.DataFrame,
            dist_mat: np.array,
            gene_names: list[str],
            t_list: Iterable[int] = (3, 3, 3),
            label_genes=(),
            dims: int = 5,
            k: int = 10,
            quantile: float = 0.02,
            max_t: int = 20,
    ):
        """
        An interactive widget for optimizing the parameters of extract_gene_trajectory

        :param gene_embedding: Gene embedding
        :param dist_mat: Gene-gene Wasserstein distance matrix (symmetric)
        :param gene_names:
        :param t_list:  Number of diffusion times to retrieve each trajectory
        :param dims: Dimensions of gene embedding to use to identify terminal genes (extrema)
        :param label_genes: Gene labels to plot (default: ())
        :param k: Adaptive kernel bandwidth for each point set to be the distance to its `K`-th nearest neighbor
        :param quantile: Thresholding parameter to extract genes for each trajectory. Default: 0.02
        :param max_t: Maximum value for t sliders. Default: 20
        :return: A widget
        """
        super().__init__()
        self.max_t = max_t

        def build_plot(dims: int, k: int, quantile: float, label_genes: list[str], **kwargs):
            tl = [v for n, v in kwargs.items() if n.startswith("_t_")]
            plt.figure(2, tight_layout=True, figsize=(6, 4))
            self.gene_trajectory = extract_gene_trajectory(gene_embedding, dist_mat, gene_names=gene_names,
                                                           t_list=tl, dims=dims, k=k, quantile=quantile)
            plot_gene_trajectory_3d(self.gene_trajectory, label_genes=label_genes)
            plt.show()

        self.f = build_plot

        self.build_widgets(t_list, dims, k, quantile, label_genes, gene_names)
        self.build_ui()

    def build_ui(self):
        def add_trajectory(_):
            self.build_widgets(self.tlist + [3], self.dims, self.k, self.quantile)
            self.build_ui()

        def remove_trajectory(_):
            self.build_widgets(self.tlist[:-1], self.dims, self.k, self.quantile)
            self.build_ui()

        add_btn = Button(text="", icon="plus")
        add_btn.on_click(add_trajectory)
        remove_btn = Button(text="", icon="minus")
        remove_btn.on_click(remove_trajectory)

        self.option_panel = VBox(children=[
            Label("Extract gene trajectories options"),
            self.dims_widget, self.k_widget, self.quantile_widget,
            Label("t_list"), *self.t_list_widgets, HBox([add_btn, remove_btn]),
            Label("genes"), self.label_genes,
        ])

        self.output = interactive_output(self.f, self.widgets)
        self.output.layout.height = '400px'

        self.children = (self.option_panel, self.output)

    def build_widgets(self, t_list: Iterable[int], dims: int, k: int, quantile: float, label_genes: list[str],
                      gene_names: list[str]):
        self.t_list_widgets = [IntSlider(t, 1, self.max_t, description=f"Trajectory-{i + 1}")
                               for i, t in enumerate(t_list)]
        self.dims_widget = IntSlider(dims, 2, 20, description="dims")
        self.k_widget = IntSlider(k, 2, 10, description="k")
        self.quantile_widget = FloatSlider(value=quantile, min=0.001, max=0.1, step=0.001, description="quantile")
        self.label_genes = TagsInput(value=list(label_genes), allowed_tags=sorted(gene_names),
                                     allow_duplicates=False, layout=Layout(width='300px'))

        self.widgets = {"dims": self.dims_widget, "k": self.k_widget,
                        "quantile": self.quantile_widget, "label_genes": self.label_genes,
                        **{f"_t_{i}": t for i, t in enumerate(self.t_list_widgets)}}



from typing import Union
import numpy as np

from gene_trajectory.util.input_validation import validate_matrix


def diffusion_map(
        dist_mat: np.array,
        k: int = 10,
        sigma: Union[float, np.array, list] = None,
        n_ev: int = 30,
        t: int = 1,
) -> tuple[np.array, np.array]:
    """
    Run a Diffusion Map

    :param dist_mat: Precomputed distance matrix (symmetric)
    :param k: Adaptive kernel bandwidth for each point set to be the distance to its `K`-th nearest neighbor
    :param sigma: Fixed kernel bandwidth, `sigma` will be ignored if `K` is specified
    :param n_ev: Number of leading eigenvectors to export
    :param t: Number of diffusion times
    :return: the diffusion embedding and the eigenvalues
    """
    validate_matrix(dist_mat, square=True)

    affinity_matrix_symm = get_symmetrized_affinity_matrix(dist_mat=dist_mat, k=k, sigma=sigma)
    normalized_vec = np.sqrt(1 / affinity_matrix_symm.sum(axis=1))
    affinity_matrix_norm = (affinity_matrix_symm * normalized_vec * normalized_vec[:, None])

    n_ev = min(n_ev, affinity_matrix_norm.shape[0])
    eigs = np.linalg.eigh(affinity_matrix_norm, )

    dm = eigs.eigenvectors * normalized_vec[:, None] * eigs.eigenvalues ** t

    diffu_emb = dm[:, :-n_ev - 1:-1]
    eigen_vals = eigs.eigenvalues[:-n_ev - 1:-1]

    return diffu_emb, eigen_vals

# Notes:
# scipy.linalg.eigh(affinity_matrix_norm, subset_by_index=[n - n_ev, n - 1]) could also be used but is 2x slower


def get_symmetrized_affinity_matrix(
        dist_mat: np.array,
        k: int = 10,
        sigma: Union[float, np.array, list] = None,
):
    """
    Computes the symmetrized distance matrix
    :param dist_mat: Precomputed distance matrix (symmetric)
    :param k: Adaptive kernel bandwidth for each point set to be the distance to its `K`-th nearest neighbor
    :param sigma: Fixed kernel bandwidth, `sigma` will be ignored if `k` is specified

    :return:
    """
    validate_matrix(dist_mat, square=True)

    dists = np.nan_to_num(dist_mat, 1e-6) # noqa
    k = min(k, dist_mat.shape[0])

    if sigma is None:
        sigma = np.apply_along_axis(func1d=sorted, axis=1, arr=dists)[:, k - 1]  # noqa
    elif np.isscalar(sigma):
        sigma = np.full(dists.shape[0], sigma)

    affinity_matrix = np.exp(-dists ** 2 / (sigma ** 2)[:, None])
    affinity_matrix_symm = (affinity_matrix + affinity_matrix.T) / 2
    return affinity_matrix_symm



from typing import Optional

import numpy as np


def validate_not_none(obj, obj_name: str = 'input'):
    if obj is None:
        raise ValueError(f"{obj_name} is None")


def validate_matrix(
        m: np.array,
        obj_name: str = 'input',
        nrows: Optional[int] = None,
        ncols: Optional[int] = None,
        shape: Optional[tuple[int, int]] = None,
        square: Optional[bool] = None,
        min_size: Optional[int] = 1,
        min_value: Optional = None,
        max_value: Optional = None,
):
    """
    Validates an input matrix
    @param m: the input matrix
    @param obj_name: the name of the object, for error reporting
    @param min_size: Minimum matrix size in each dimension. Defaults to 1, and will raise for an empty matrix
    @param nrows: Number of rows in the matrix
    @param ncols: Number of rows in the matrix
    @param shape: the expected shape of the matrix
    @param square: If True, ensures the matrix is square. If False, ensures the matrix is not
    @param min_value: Minimum value for each element
    @param max_value: Maximum value for each element
    """
    validate_not_none(m, obj_name=obj_name)

    if len(m.shape) != 2:
        raise ValueError(f"{obj_name} is not a matrix. Shape: {m.shape}")
    mr, mc = m.shape

    if nrows is not None:
        if mr != nrows:
            raise ValueError(f"{obj_name} does not have {nrows} rows. Shape: {m.shape}")

    if ncols is not None:
        if mc != ncols:
            raise ValueError(f"{obj_name} does not have {ncols} columns. Shape: {m.shape}")

    if shape is not None:
        if m.shape != shape:
            raise ValueError(f"{obj_name} does not have shape {shape}. Shape: {m.shape}")

    if square is True:
        if mr != mc:
            raise ValueError(f"{obj_name} is not a square matrix. Shape: {m.shape}")
    elif square is False:
        if mr == mc:
            raise ValueError(f"{obj_name} is a square matrix. Shape: {m.shape}")

    if min_size is not None:
        for s in m.shape:
            if s < min_size:
                raise ValueError(f"{obj_name} does not have enough elements. Min_size: {min_size}, Shape: {m.shape}")

    if min_value is not None:
        if m.min() < min_value:
            raise ValueError(f"{obj_name} should not have values less than {min_value}. Minimum found: {m.min()}")

    if max_value is not None:
        if m.max() > max_value:
            raise ValueError(f"{obj_name} should not have values greater than {max_value}. Maximum found: {m.max()}")



import argparse
import os
import scipy.io as sio
from scipy.sparse import issparse
import numpy as np

from gene_trajectory.gene_distance_shared import _DEFAULT_NUMITERMAX
from gene_trajectory.gene_distance_shared import cal_ot_mat as cal_ot_mat_from_numpy


def parse_args():
    parser = argparse.ArgumentParser(description='Calculate the earth mover distance matrix')
    parser.add_argument('--path', type=os.path.abspath, required=True,
                        help='The path containing the cost matrix (ot_cost.csv)' +
                             ' and the gene expression matrix (gene_expression.mtx)')
    parser.add_argument('--num_iter_max', type=int, default=_DEFAULT_NUMITERMAX,
                        help='The max number of iterations when computing the distance (see ot.emd2)')
    parser.add_argument('--show_progress_bar', type=bool, default=True,
                        help='Shows a progress bar while running the computation (default: True)')
    parser.add_argument('--processes', type=int, required=False, default=None,
                        help='The number of processes to use (defaults to the number of CPUs available)')
    return parser.parse_args()


def cal_ot_mat(
    path: str,
    num_iter_max=_DEFAULT_NUMITERMAX,
    show_progress_bar=True,
    processes: int = None,
) -> None:
    """
    Calculate the earth mover distance matrix. Note that this step is computationally expensive
    and will be performed in parallel.

    :param path: path to the folder where the cost matrix (ot_cost.csv) and the gene expression matrix
                 (gene_expression.mtx) are saved
    :param num_iter_max: the max number of iterations when computing the distance (see ot.emd2)
    :param show_progress_bar: shows a progress bar while running the computation (default: True)
    :param processes:the number of processes to use (defaults to the number of CPUs available)
    """
    ot_cost = np.loadtxt(os.path.join(path, "ot_cost.csv"), delimiter=",")
    gene_expr = sio.mmread(os.path.join(path,"gene_expression.mtx"))
    if issparse(gene_expr):
        gene_expr = gene_expr.todense()
    gene_pairs_file = os.path.join(path, "gene_pairs.csv")
    gene_pairs = np.loadtxt(gene_pairs_file, delimiter=",").astype(int) if os.path.isfile(gene_pairs_file) else None

    emd_mat2 = cal_ot_mat_from_numpy(ot_cost=ot_cost, gene_expr=gene_expr,
                                     gene_pairs=gene_pairs, num_iter_max=num_iter_max,
                                     show_progress_bar=show_progress_bar, processes=processes)
    np.savetxt(path + "/emd.csv", emd_mat2, delimiter=",")


if __name__ == '__main__':
    args = parse_args()

    cal_ot_mat(path=args.path,
               num_iter_max=args.num_iter_max,
               show_progress_bar=args.show_progress_bar,
               processes=args.processes)



import pandas as pd
import scanpy as sc
from matplotlib import pyplot as plt
import seaborn as sns
from typing import Iterable, Any


def plot_gene_trajectory_3d(
        gene_trajectory: pd.DataFrame,
        s: int = 1,
        label_genes: Iterable[str] = None,
        **kwargs: Any
) -> None:
    """
    Generate a 3D plot of a gene-trajectory object

    :param gene_trajectory: a gene trajectory result
    :param s: scatterplot point size (default: 1)
    :param label_genes: Gene labels to plot (default: None)
    :param kwargs: plot arguments that will be passed to Axes.scatter
    """
    for c in 'DM_1', 'DM_2', 'DM_3', 'selected':
        if c not in gene_trajectory.columns:
            raise ValueError(f'Column {c} is not present in gene trajectory DataFrame')

    ax = plt.axes(projection='3d')
    selections = pd.Categorical(gene_trajectory.selected)

    for c in selections.categories:
        idxs = selections == c
        ax.scatter(xs=gene_trajectory['DM_1'][idxs],
                   ys=gene_trajectory['DM_2'][idxs],
                   zs=gene_trajectory['DM_3'][idxs],
                   s=s,
                   label=c,
                   **kwargs)

    if label_genes:
        for g in label_genes:
            ax.text(x=gene_trajectory['DM_1'][g],
                    y=gene_trajectory['DM_2'][g],
                    z=gene_trajectory['DM_3'][g],
                    s=g)
    ax.legend()


def plot_gene_trajectory_2d(
        gene_trajectory: pd.DataFrame,
        s: int = 1,
        label_genes: Iterable[str] = None,
        **kwargs: Any
) -> None:
    """
    Generate a 2D plot of a gene-trajectory object

    :param gene_trajectory: a gene trajectory result
    :param s: scatterplot point size (default: 1)
    :param label_genes: Gene labels to plot (default: None)
    :param kwargs: plot arguments that will be passed to sns.scatterplot
    """
    for c in 'DM_1', 'DM_2', 'selected':
        if c not in gene_trajectory.columns:
            raise ValueError(f'Column {c} is not present in gene trajectory DataFrame')
    sns.scatterplot(data=gene_trajectory,
                    x='DM_1',
                    y='DM_2',
                    hue='selected',
                    s=s,
                    **kwargs)
    if label_genes:
        ax = plt.gca()
        for g in label_genes:
            ax.text(x=gene_trajectory['DM_1'][g],
                    y=gene_trajectory['DM_2'][g],
                    s=g)


def plot_gene_trajectory_umap(
        adata: sc.AnnData,
        trajectory: str = 'Trajectory1',
        other_panels: Iterable[str] = (),
        reverse: bool = False,
        cmap: str = 'RdYlBu_r',
        **kwargs: Any,
) -> None:
    """
    Generate a series of umap plot for gene trajectory bins

    :param adata: a dataset with gene trajectory metadata
    :param trajectory: the name of the trajectory
    :param other_panels: other panels to be added to the umap
    :param reverse: reverse the order of the panels
    :param cmap: colormap to be used
    :param kwargs: plot arguments that will be passed to scanpy.pl.umap
    """
    other_panels = [other_panels] if isinstance(other_panels, str) else other_panels
    trajectory_panels = [k for k in adata.obs_keys() if k.startswith(trajectory)]
    if not trajectory_panels:
        raise ValueError(f'No obj metadata found for {trajectory}')
    if reverse:
        trajectory_panels.reverse()
    panels = [*trajectory_panels, *other_panels]
    sc.pl.umap(adata, color=panels, cmap=cmap, **kwargs)



import scanpy as sc
from sklearn.metrics import pairwise_distances

from gene_trajectory.diffusion_map import diffusion_map
import logging

logger = logging.getLogger()


def run_dm(
        adata: sc.AnnData,
        reduction: str = "X_pca",
        k=10,
        sigma=None,
        n_components=30,
        t=1,
        dist_mat=None,
        reduction_result="X_dm",
) -> None:
    """
    Run Diffusion Map on a Scanpy Anndata object

    :param adata: Scanpy Anndata object
    :param reduction: Dimensionality reduction to use, default: 'X_pca'
    :param k: Adaptive kernel bandwidth for each point set to be the distance to its `K`-th nearest neighbor.
    :param sigma: Fixed kernel bandwidth, `sigma` will be ignored if `K` is specified.
    :param n_components: Number of leading nontrivial eigenvectors to export
    :param t: Number of diffusion times
    :param dist_mat: Precomputed distance matrix (optional)
    :param reduction_result: Dimensionality reduction to store result, default: 'X_dm'
    """
    if reduction == "X_pca" and reduction not in adata.obsm_keys():
        logger.info('Computing PCA')
        sc.pp.pca(adata)

    if reduction not in adata.obsm_keys():
        raise ValueError(f'Reduction "{reduction}" is not present. Available: {adata.obsm_keys()}')

    if dist_mat is None:
        dist_mat = pairwise_distances(adata.obsm[reduction])

    diffu_emb, _ = diffusion_map(dist_mat, k=k, sigma=sigma, n_ev=n_components + 1, t=t)
    diffu_emb = diffu_emb[:, 1:]  # remove the first eigenvector (trivial)
    adata.obsm[reduction_result] = diffu_emb



import logging
from urllib.request import urlretrieve
import hashlib
import os

logger = logging.getLogger()


def download_file_if_missing(
        file: str,
        url: str,
        create_target_folder_if_missing: bool = False,
        md5_hash: str = None,
):
    """
    Checks if a file exists and downloads it if missing

    :param file: where the file should be saved
    :param url: the url to download from
    :param create_target_folder_if_missing: creates the folder where the target file should be saved if not existing
    :param md5_hash: the md5 hash of the file
    """
    outdir = os.path.dirname(file)
    if outdir and not os.path.isdir(outdir):
        if create_target_folder_if_missing:
            os.makedirs(outdir, exist_ok=True)
        else:
            raise ValueError(f'Directory does not exist: {outdir}')

    if not os.path.exists(file):
        print(f'\nDownloading {url} to {file}...')
        urlretrieve(url, file)
        print(f'File saved on {file}')
    if md5_hash:
        check_md5_hash(file, md5_hash)


def check_md5_hash(file: str, md5_hash: str):
    with open(file, "rb") as f:
        hexdigest = hashlib.file_digest(f, "md5").hexdigest() # noqa
    if hexdigest != md5_hash:
        raise ValueError(f'The md5 hash of {file} is {hexdigest} and does not match {md5_hash}')



import logging
from concurrent.futures import ThreadPoolExecutor
import os
import time
from multiprocessing.managers import SharedMemoryManager
from typing import Optional, Sized

import numpy as np
import ot
from tqdm import tqdm

from gene_trajectory.util.input_validation import validate_matrix
from gene_trajectory.util.shared_array import SharedArray, PartialStarApply

logger = logging.getLogger()
_DEFAULT_NUMITERMAX = 50000

# Implementation notes:
# - The matrices can get relatively large so they are kept in memory as shared objects
# - Using ThreadPoolExecutor as ProcessPoolExecutor (which would make more sense) sometimes halts on Mac,
#   even when changing multiprocessing.set_start_method('fork')


def cal_ot_mat(
        ot_cost: np.array,
        gene_expr: np.array,
        gene_pairs: Optional[Sized] = None,
        num_iter_max=_DEFAULT_NUMITERMAX,
        show_progress_bar=True,
        processes: Optional[int] = None,
) -> np.array:
    """
    Calculate the earth mover distance matrix. Note that this step is computationally expensive
    and will be performed in parallel.

    :param ot_cost: the cost matrix
    :param gene_expr: the gene expression matrix
    :param gene_pairs: only compute the distance for the given pairs (0-indexed) (default: None).
           the distance entry for missing pairs will be set to 1000*max(computed_gene_distances)
    :param num_iter_max: the max number of iterations when computing the distance (see ot.emd2)
    :param show_progress_bar: shows a progress bar while running the computation (default: True)
    :param processes:the number of processes to use (defaults to the number of CPUs available)
    :return: the distance matrix
    """
    processes = int(processes) if isinstance(processes, float) else os.cpu_count()
    validate_matrix(gene_expr, obj_name='Gene Expression Matrix', min_value=0)
    ncells, ngenes = gene_expr.shape
    validate_matrix(ot_cost, obj_name='Cost Matrix', shape=(ncells, ncells), min_value=0)

    if show_progress_bar:
        logger.info(f'Computing emd distance..')

    if gene_pairs is None:
        pairs = ((i, j) for i in range(0, ngenes - 1) for j in range(i + 1, ngenes))
        npairs = (ngenes * (ngenes - 1)) // 2
    else:
        pairs = gene_pairs
        npairs = len(gene_pairs)
    emd_mat = np.full((ngenes, ngenes), fill_value=np.NaN)

    with SharedMemoryManager() as manager:
        start_time = time.perf_counter()
        # create and configure the process pool

        with ThreadPoolExecutor(max_workers=processes) as pool:
            # prepare shared environment
            cost = SharedArray.copy(manager, np.asarray(ot_cost))
            gexp = SharedArray.copy(manager, np.asarray(gene_expr))
            f = PartialStarApply(_cal_ot, cost, gexp)

            # execute tasks and process results
            result_generator = pool.map(f,  ((num_iter_max, i, j) for i, j in pairs))
            if show_progress_bar:
                result_generator = tqdm(result_generator, total=npairs, position=0, leave=True)
            for d, i, j in result_generator:
                emd_mat[i, j] = emd_mat[j, i] = d
        finish_time = time.perf_counter()
        if show_progress_bar:
            logger.info("Program finished in {} seconds - using multiprocessing".format(finish_time - start_time))

        np.fill_diagonal(emd_mat, 0)
        np.nan_to_num(emd_mat, nan=1000 * np.nanmax(emd_mat), copy=False)

        return emd_mat


def _cal_ot(ot_cost_matrix: np.array, gene_expr_matrix: np.array, num_iter_max: int, i: int, j: int):
    """
    Computes the EMD between i and j
    """
    gene_i = np.array(gene_expr_matrix[:, i]).reshape(-1)
    gene_j = np.array(gene_expr_matrix[:, j]).reshape(-1)
    gene_i = gene_i / sum(gene_i)
    gene_j = gene_j / sum(gene_j)

    emd_dist = ot.emd2(gene_i[np.nonzero(gene_i)],
                       gene_j[np.nonzero(gene_j)],
                       ot_cost_matrix[np.nonzero(gene_i)[0], :][:, np.nonzero(gene_j)[0]],
                       numItermax=num_iter_max)
    # return the generated value
    return emd_dist, i, j



from typing import Optional
import numpy as np
import pandas as pd
import scanpy as sc


def add_gene_bin_score(
        adata: sc.AnnData,
        gene_trajectory: pd.DataFrame,
        n_bins: int = 5,
        trajectories: int = 2,
        layer: Optional[str] = None,
        reverse: Optional[bool] = None,
        prefix: str = 'Trajectory',
) -> None:
    """
    Add gene bin score

    :param adata a scanpy Anndata object
    :param gene_trajectory Gene trajectory data frame
    :param n_bins How many gene bins
    :param trajectories: Which gene trajectories to define gene bin score
    :param layer string Which layer to use. Uses adata.X if empty
    :param reverse: Whether to reverse the order of genes along each trajectory
    :param prefix: String added to the names in the metadata
    """

    if layer is not None and layer not in adata.layers.keys():
        raise ValueError(f'Layer {layer} not found in adata. Available {list(adata.layers.keys())}')

    for trajectory in range(1, trajectories + 1):
        trajectory_name = f'Pseudoorder-{trajectory}'

        gene_trajectory_reordered = gene_trajectory.sort_values(trajectory_name)
        genes = gene_trajectory_reordered[gene_trajectory_reordered[trajectory_name] > 0].index.values
        if reverse:
            genes = list(reversed(genes))
        step = len(genes) / n_bins

        for i in range(n_bins):
            start = int(np.ceil(i * step))
            end = min(int(np.ceil((i + 1) * step)), len(genes))
            genes_subset = genes[start:end]

            adata_subset = adata[:, genes_subset]
            x = adata_subset.layers[layer] if layer else adata_subset.X
            normalized_gc = x > 0
            meta = np.squeeze(np.asarray(normalized_gc.sum(axis=1) / len(genes_subset)))
            adata.obs[f'{prefix}{trajectory}_genes{i + 1}'] = meta
